@@ -45,6 +45,7 @@ unsafe fn get_isr_yielder() -> Option<unsafe fn()> {
 pub unsafe fn set_isr_yielder(yielder: Option<unsafe fn()>) -> Option<unsafe fn()> {
     if active() {
         let ptr = if let Some(yielder) = yielder {
+            #[allow(clippy::transmutes_expressible_as_ptr_casts)]
             core::mem::transmute(yielder)
         } else {
             ptr::null_mut()
@@ -477,3 +478,25 @@ impl<T> embedded_svc::mutex::Mutex for Mutex<T> {
         Mutex::lock(self)
     }
 }
+
+#[cfg(all(feature = "experimental", feature = "embedded-svc"))]
+pub struct MutexSignalFamily;
+
+#[cfg(all(feature = "experimental", feature = "embedded-svc"))]
+impl embedded_svc::signal::asyncs::SignalFamily for MutexSignalFamily {
+    type Signal<T> = embedded_svc::utils::asyncs::signal::MutexSignal<
+        Mutex<embedded_svc::utils::asyncs::signal::State<T>>,
+        T,
+    >;
+}
+
+#[cfg(all(feature = "experimental", feature = "embedded-svc"))]
+impl embedded_svc::signal::asyncs::SendSyncSignalFamily for MutexSignalFamily {
+    type Signal<T>
+    where
+        T: Send,
+    = embedded_svc::utils::asyncs::signal::MutexSignal<
+        Mutex<embedded_svc::utils::asyncs::signal::State<T>>,
+        T,
+    >;
+}
