@@ -15,6 +15,8 @@ use crate::i2c;
 #[cfg(not(feature = "riscv-ulp-hal"))]
 use crate::ledc;
 #[cfg(not(feature = "riscv-ulp-hal"))]
+use crate::rmt;
+#[cfg(not(feature = "riscv-ulp-hal"))]
 use crate::serial;
 #[cfg(not(feature = "riscv-ulp-hal"))]
 use crate::spi;
@@ -47,6 +49,8 @@ pub struct Peripherals {
     pub can: can::CAN,
     #[cfg(not(feature = "riscv-ulp-hal"))]
     pub ledc: ledc::Peripheral,
+    #[cfg(not(feature = "riscv-ulp-hal"))]
+    pub rmt: rmt::Peripheral,
     #[cfg(all(any(esp32, esp32s2, esp32s3), not(feature = "riscv-ulp-hal")))]
     pub ulp: ulp::ULP,
 }
@@ -95,6 +99,8 @@ impl Peripherals {
             can: can::CAN::new(),
             #[cfg(not(feature = "riscv-ulp-hal"))]
             ledc: ledc::Peripheral::new(),
+            #[cfg(not(feature = "riscv-ulp-hal"))]
+            rmt: rmt::Peripheral::new(),
             #[cfg(all(any(esp32, esp32s2, esp32s3), not(feature = "riscv-ulp-hal")))]
             ulp: ulp::ULP::new(),
         }
